from .base_client import APIError, ResultsPage

# Note that we import DetectorPlatformClient thrice, to export it under two names:
# - DetectorPlatformClient as the name it should be used with
# - APIClient and DetectorPlatformClient to preserve backward compatibility, since that was the name it was
# exported under previously (when we originally had only one platform and API client).
# Ditto for PlotsAnalysisPlatformClient / TracerClient
from .forge_client import ForgeClient
from .forge_client import ForgeClient as APIClient
from .forge_client import ForgeClient as DetectorPlatformClient
from .nongeo import nongeo_result_to_pixel
from .tracer_client import TracerClient
from .tracer_client import TracerClient as PlotsAnalysisPlatformClient

__all__ = ["APIClient", "DetectorPlatformClient", "ForgeClient", "PlotsAnalysisPlatformClient", "TracerClient", "nongeo_result_to_pixel", "APIError", "ResultsPage"]



from __future__ import annotations

import json
import logging
import os
import sys
import time
from collections.abc import Callable, Iterator

if sys.version_info >= (3, 8):
    from typing import Literal, TypedDict
else:
    from typing_extensions import Literal, TypedDict

from typing import Any, Generic, TypeVar
from urllib.parse import urlencode, urljoin

import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

logger = logging.getLogger()

CHUNK_SIZE_BYTES = 8192  # 8 KiB


# allow injecting an non-existing package name to test the fallback behavior
# of _get_ua in tests (see test_headers_user_agent_version__fallback)
def _get_distr_name():
    return 'picterra'


def _get_ua():
    import platform
    pkg = _get_distr_name()
    if sys.version_info >= (3, 8):
        from importlib.metadata import PackageNotFoundError, version
        try:
            ver = version(pkg)
        except PackageNotFoundError:
            ver = 'no_version'
    else:
        import pkg_resources  # type: ignore[import]
        try:
            ver = pkg_resources.require(pkg)[0].version
        except pkg_resources.DistributionNotFound:
            ver = 'no_version'
    o_s = " ".join([os.name, platform.system(), platform.release()])
    v_info = sys.version_info
    py = "Python " + str(v_info.major) + "." + str(v_info.minor)
    return "picterra-python/%s (%s %s)" % (ver, py, o_s,)


class APIError(Exception):
    """Generic API error exception"""

    pass


class _RequestsSession(requests.Session):
    """
    Override requests session to to implement a global session timeout
    """

    def __init__(self, *args, **kwargs):
        self.timeout = kwargs.pop("timeout")
        super().__init__(*args, **kwargs)
        self.headers.update(
            {
                "User-Agent": "%s - %s" % (_get_ua(), self.headers["User-Agent"])
            }
        )

    def request(self, *args, **kwargs):
        kwargs.setdefault("timeout", self.timeout)
        return super().request(*args, **kwargs)


def _download_to_file(url: str, filename: str):
    # Given we do not use self.sess the timeout is disabled (requests default), and this
    # is good as file download can take a long time
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(filename, "wb+") as f:
            logger.debug("Downloading to file %s.." % filename)
            for chunk in r.iter_content(chunk_size=CHUNK_SIZE_BYTES):
                if chunk:  # filter out keep-alive new chunks
                    f.write(chunk)


def _upload_file_to_blobstore(upload_url: str, filename: str):
    if not (os.path.exists(filename) and os.path.isfile(filename)):
        raise ValueError("Invalid file: " + filename)
    with open(
        filename, "rb"
    ) as f:  # binary recommended by requests stream upload (see link below)
        logger.debug("Opening and streaming to upload file %s" % filename)
        # Given we do not use self.sess the timeout is disabled (requests default), and this
        # is good as file upload can take a long time. Also we use requests streaming upload
        # (https://requests.readthedocs.io/en/latest/user/advanced/#streaming-uploads) to avoid
        # reading the (potentially large) layer GeoJSON in memory
        resp = requests.put(upload_url, data=f)
    if not resp.ok:
        logger.error("Error when uploading to blobstore %s" % upload_url)
        raise APIError(resp.text)


def multipolygon_to_polygon_feature_collection(mp):
    return {
        "type": "FeatureCollection",
        "features": [{
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "Polygon",
                "coordinates": p
            }
        } for p in mp["coordinates"]]
    }


def _check_resp_is_ok(resp: requests.Response, msg: str) -> None:
    if not resp.ok:
        raise APIError("%s (status %d): %s" % (msg, resp.status_code, resp.text))


T = TypeVar("T")


class ResultsPage(Generic[T]):
    """
    Interface for a paginated response from the API

    Typically the endpoint returning list of objects return them splitted
    in pages (page 1, page 2, etc..) of a fixed dimension (eg 20). Thus
    each `list_XX` function returns a ResultsPage (by default the first one);
    once you have a ResultsPage for a given list of objects, you can:

    * check its length with ``len()``;
        - example: ``len(page)``
    * access a single element with the index operator ``[]``;
        - example: ``page[5]``
    * turn it into a list of dictionaries with ``list()``;
        - example: ``list(page)``
    * get the next page with ``.next()``; this could return None if the list is finished;
        - example: ``page.next()``

    You can also get a specific page passing the page number to the ``list_XX`` function
    """
    _fetch: Callable[[str], requests.Response]
    _next_url: str | None
    _prev_url: str | None
    _results: list[T]
    _url: str

    def __init__(self, url: str, fetch: Callable[[str], requests.Response]):
        resp = fetch(url)
        _check_resp_is_ok(resp, "Failed to get page")
        r: dict[str, Any] = resp.json()
        next_url: str | None = r["next"]
        prev_url: str | None = r["previous"]
        results: list[T] = r["results"]

        self._fetch = fetch
        self._next_url = next_url
        self._prev_url = prev_url
        self._results = results
        self._url = url

    def next(self):
        return ResultsPage(self._next_url, self._fetch) if self._next_url else None

    def previous(self):
        return ResultsPage(self._prev_url, self._fetch) if self._prev_url else None

    def __len__(self) -> int:
        return len(self._results)

    def __getitem__(self, key: int) -> T:
        return self._results[key]

    def __iter__(self) -> Iterator[T]:
        return iter([self._results[i] for i in range(len(self._results))])

    def __str__(self) -> str:
        return f"{len(self._results)} results from {self._url}"


class Feature(TypedDict):
    type: Literal["Feature"]
    properties: dict[str, Any]
    geometry: dict[str, Any]


class FeatureCollection(TypedDict):
    type: Literal["FeatureCollection"]
    features: list[Feature]


class BaseAPIClient:
    """
    Base class for Picterra API clients.

    This is subclassed for the different products we have.
    """

    def __init__(
        self, api_url: str, timeout: int = 30, max_retries: int = 3, backoff_factor: int = 10
    ):
        """
        Args:
            api_url: the api's base url. This is different based on the Picterra product used
                and is typically defined by implementations of this client
            timeout: number of seconds before the request times out
            max_retries: max attempts when ecountering gateway issues or throttles; see
                retry_strategy comment below
            backoff_factor: factor used nin the backoff algorithm; see retry_strategy comment below
        """
        base_url = os.environ.get(
            "PICTERRA_BASE_URL", "https://app.picterra.ch/"
        )
        api_key = os.environ.get("PICTERRA_API_KEY", None)
        if not api_key:
            raise APIError("PICTERRA_API_KEY environment variable is not defined")
        logger.info(
            "Using base_url=%s, api_url=%s; %d max retries, %d backoff and %s timeout.",
            base_url,
            api_url,
            max_retries,
            backoff_factor,
            timeout,
        )
        self.base_url = urljoin(base_url, api_url)
        # Create the session with a default timeout (30 sec), that we can then
        # override on a per-endpoint basis (will be disabled for file uploads and downloads)
        self.sess = _RequestsSession(timeout=timeout)
        # Retry: we set the HTTP codes for our throttle (429) plus possible gateway problems (50*),
        # and for polling methods (GET), as non-idempotent ones should be addressed via idempotency
        # key mechanism; given the algorithm is {<backoff_factor> * (2 **<retries-1>}, and we
        # default to 30s for polling and max 30 req/min, the default 5-10-20 sequence should
        # provide enough room for recovery
        retry_strategy = Retry(
            total=max_retries,
            status_forcelist=[429, 502, 503, 504],
            backoff_factor=backoff_factor,
            allowed_methods=["GET"],
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        self.sess.mount("https://", adapter)
        self.sess.mount("http://", adapter)
        # Authentication
        self.sess.headers.update({"X-Api-Key": api_key})

    def _full_url(self, path: str, params: dict[str, Any] | None = None):
        url = urljoin(self.base_url, path)
        if not params:
            return url
        else:
            qstr = urlencode(params)
            return "%s?%s" % (url, qstr)

    def _wait_until_operation_completes(
        self, operation_response: dict[str, Any]
    ) -> dict[str, Any]:
        """Polls an operation an returns its data"""
        operation_id = operation_response["operation_id"]
        poll_interval = operation_response["poll_interval"]
        # Just sleep for a short while the first time
        time.sleep(poll_interval * 0.1)
        while True:
            logger.info("Polling operation id %s" % operation_id)
            resp = self.sess.get(
                self._full_url("operations/%s/" % operation_id),
            )
            if not resp.ok:
                raise APIError(resp.text)
            status = resp.json()["status"]
            logger.info("status=%s" % status)
            if status == "success":
                break
            if status == "failed":
                errors = resp.json()["errors"]
                raise APIError(
                    "Operation %s failed: %s" % (operation_id, json.dumps(errors))
                )
            time.sleep(poll_interval)
        return resp.json()

    def _return_results_page(
        self, resource_endpoint: str, params: dict[str, Any] | None = None
    ) -> ResultsPage:
        if params is None:
            params = {}
        if "page_number" not in params:
            params["page_number"] = 1

        url = self._full_url("%s/" % resource_endpoint, params=params)
        return ResultsPage(url, self.sess.get)

    def get_operation_results(self, operation_id: str) -> dict[str, Any]:
        """
        Return the 'results' dict of an operation

        This a **beta** function, subject to change.

        Args:
            operation_id: The id of the operation
        """
        resp = self.sess.get(
            self._full_url("operations/%s/" % operation_id),
        )
        return resp.json()["results"]



import json
import re
import time

import httpretty
import pytest
import requests
import responses

from picterra import base_client
from picterra.forge_client import ForgeClient
from tests.utils import _add_api_response, _client, detector_api_url


def test_forge_client_base_url(monkeypatch):
    """
    Sanity-check that the client defaults to the correct base url
    """
    monkeypatch.setenv("PICTERRA_API_KEY", "1234")
    client = ForgeClient()
    assert client.base_url == "https://app.picterra.ch/public/api/v2/"


# Cannot test Retry with responses, @see https://github.com/getsentry/responses/issues/135
@httpretty.activate
def test_backoff_success(monkeypatch):
    data = {"count": 0, "next": None, "previous": None, "results": []}
    httpretty.register_uri(
        httpretty.GET,
        detector_api_url("rasters/"),
        responses=[
            httpretty.Response(body=None, status=429),
            httpretty.Response(body=None, status=502),
            httpretty.Response(body=json.dumps(data), status=200),
        ],
    )
    client = _client(monkeypatch, max_retries=2, backoff_factor=0.1)
    client.list_rasters()
    assert len(httpretty.latest_requests()) == 3


@httpretty.activate
def test_backoff_failure(monkeypatch):
    httpretty.register_uri(
        httpretty.GET,
        detector_api_url("rasters/"),
        responses=[
            httpretty.Response(
                body=None,
                status=429,
            ),
            httpretty.Response(body=None, status=502),
            httpretty.Response(body=None, status=502),
        ],
    )
    client = _client(monkeypatch, max_retries=1)
    with pytest.raises(requests.exceptions.ConnectionError):
        client.list_rasters()
    assert len(httpretty.latest_requests()) == 2


@httpretty.activate
def test_timeout(monkeypatch):
    def request_callback(request, uri, response_headers):
        time.sleep(2)
        return [200, response_headers, json.dumps([])]

    httpretty.register_uri(httpretty.GET, detector_api_url("rasters/"), body=request_callback)
    timeout = 1
    client = _client(monkeypatch, timeout=timeout)
    with pytest.raises(requests.exceptions.ConnectionError) as e:
        client.list_rasters()
    full_error = str(e.value)
    assert "MaxRetryError" not in full_error
    assert "timeout" in full_error
    assert "read timeout=%d" % timeout in full_error
    assert len(httpretty.latest_requests()) == 1


@responses.activate
def test_headers_user_agent_version(monkeypatch):
    _add_api_response(detector_api_url("detectors/"), responses.POST, json={"id": "foobar"})
    client = _client(monkeypatch)
    client.create_detector()
    assert len(responses.calls) == 1
    ua = responses.calls[0].request.headers["User-Agent"]
    regex = "^picterra-python/\d+\.\d+"
    assert re.compile(regex).match(ua) is not None


@responses.activate
def test_headers_user_agent_version__fallback(monkeypatch):
    _add_api_response(detector_api_url("detectors/"), responses.POST, json={"id": "foobar"},)
    monkeypatch.setattr(base_client, '_get_distr_name', lambda: 'foobar')
    client = _client(monkeypatch)
    client.create_detector()
    assert len(responses.calls) == 1
    ua = responses.calls[0].request.headers["User-Agent"]
    regex = "^picterra-python/no_version"
    assert re.compile(regex).match(ua) is not None


@responses.activate
def test_results_page():
    responses.add(
        method=responses.GET,
        url="http://example.com/page/1",
        json={
            "count": 2,
            "next": "http://example.com/page/2",
            "previous": None,
            "results": ["one", "two"],
        },
        status=200,
    )
    responses.add(
        method=responses.GET,
        url="http://example.com/page/2",
        json={
            "count": 1, "next": None, "previous": "http://example.com/page/1",
            "results": ["three"],
        },
        status=200,
    )
    page1 = base_client.ResultsPage("http://example.com/page/1", requests.get)
    assert isinstance(page1, base_client.ResultsPage)
    assert len(page1) == 2 and page1.previous() is None
    assert page1[0] == "one" and page1[1] == "two"
    assert list(page1)[0] == "one" and list(page1)[1] == "two"
    assert str(page1) == "2 results from http://example.com/page/1"
    page2 = page1.next()
    assert str(page2) == "1 results from http://example.com/page/2"
    assert isinstance(page2, base_client.ResultsPage)
    assert len(page2) == 1 and page2[0] == "three"
    assert page2.next() is None
    assert list(page2.previous())[0] == "one"



import json
import tempfile

import pytest
import responses

from picterra.base_client import multipolygon_to_polygon_feature_collection
from tests.utils import (
    OP_RESP,
    OPERATION_ID,
    TEST_POLL_INTERVAL,
    _add_api_response,
    _client,
    detector_api_url,
)


def add_mock_rasters_list_response(endpoint=detector_api_url("rasters/")):
    data1 = {
        "count": 5,
        "next": "%s?page_number=2" % endpoint,
        "previous": None,
        "page_size": 2,
        "results": [
            {"id": "40", "status": "ready", "name": "raster1"},
            {"id": "41", "status": "ready", "name": "raster2"},
        ],
    }
    data2 = {
        "count": 5,
        "next": "%s?page_number=3" % endpoint,
        "previous": None,
        "page_size": "2",
        "results": [
            {"id": "42", "status": "ready", "name": "raster3"},
            {"id": "43", "status": "ready", "name": "raster4"},
        ],
    }
    data3 = {
        "count": 5,
        "next": None,
        "previous": "%s?page_number=2" % endpoint,
        "page_size": 2,
        "results": [
            {"id": "44", "status": "ready", "name": "raster5"},
        ],
    }
    _add_api_response(
        endpoint,
        json=data1,
        match=responses.matchers.query_param_matcher({"page_number": "1"}),
    )
    _add_api_response(
        endpoint,
        json=data2,
        match=responses.matchers.query_param_matcher({"page_number": "2"}),
    )
    _add_api_response(
        endpoint,
        json=data3,
        match=responses.matchers.query_param_matcher({"page_number": "3"}),
    )


def add_mock_rasters_in_folder_list_response(folder_id):
    data = {
        "count": 1,
        "next": None,
        "previous": None,
        "page_size": 2,
        "results": [
            {
                "id": "77",
                "status": "ready",
                "name": "raster_in_folder1",
                "folder_id": folder_id,
            },
        ],
    }
    qs = {"folder": folder_id, "page_number": "1"}
    _add_api_response(
        detector_api_url("rasters/"), json=data, match=responses.matchers.query_param_matcher(qs)
    )


def add_mock_rasters_in_filtered_list_response(
    search=None, tag=None, cloud=None, before=None, after=None, has_layers=None, page=1
):
    name = (search + "_" if search else "") + "raster" + ("_" + tag if tag else "")
    data = {
        "count": 1,
        "next": None,
        "previous": None,
        "page_size": 2,
        "results": [
            {"id": "77", "status": "ready", "name": name},
        ],
    }
    qs = {"page_number": page}
    if search is not None:
        qs["search"] = search
    if tag is not None:
        qs["user_tag"] = tag
    if cloud is not None:
        qs["max_cloud_coverage"] = cloud
    if before:
        qs["captured_before"] = before
    if after:
        qs["captured_after"] = after
    if has_layers is not None:
        qs["has_vector_layers"] = bool(has_layers)
    _add_api_response(
        detector_api_url("rasters/"), match=responses.matchers.query_param_matcher(qs), json=data
    )


def add_mock_vector_layers_filtered_list_response(
    idx, raster, search=None, detector=None
):
    name = f"layer_{idx}"
    data = {
        "count": 1,
        "next": None,
        "previous": None,
        "page_size": 1,
        "results": [{"id": str(idx), "count": idx, "name": name}],
    }
    qs = {"page_number": 1}
    if search is not None:
        qs["search"] = search
    if detector is not None:
        qs["detector"] = detector
    _add_api_response(
        detector_api_url(f"rasters/{raster}/vector_layers/"),
        match=responses.matchers.query_param_matcher(qs),
        json=data,
    )


def add_mock_detectors_list_response(string=None, tag=None, shared=None):
    data1 = {
        "count": 4,
        "next": detector_api_url("detectors/?page_number=2"),
        "previous": None,
        "page_size": 2,
        "results": [
            {"id": "40", "type": "count", "name": string or "detector1"},
            {"id": "41", "type": "count", "name": string or "detector2"},
        ],
    }
    data2 = {
        "count": 4,
        "next": None,
        "previous": detector_api_url("detectors/?page_number=1"),
        "page_size": 2,
        "results": [
            {"id": "42", "type": "count", "name": string or "detector3"},
            {"id": "43", "type": "count", "name": string or "detector4"},
        ],
    }
    qs_params = {"page_number": "1"}
    if string:
        qs_params["search"] = string
    if tag:
        qs_params["user_tag"] = tag
    if shared:
        qs_params["is_shared"] = shared
    _add_api_response(
        detector_api_url("detectors/"),
        match=responses.matchers.query_param_matcher(qs_params),
        json=data1,
    )
    qs_params2 = {"page_number": "2"}
    if string:
        qs_params2["search"] = string
    if tag:
        qs_params2["user_tag"] = tag
    if shared:
        qs_params2["is_shared"] = shared
    _add_api_response(
        detector_api_url("detectors/"),
        match=responses.matchers.query_param_matcher(qs_params2),
        json=data2,
    )


def add_mock_detector_creation_response(**kwargs):
    match = responses.json_params_matcher({"configuration": kwargs}) if kwargs else None
    _add_api_response(detector_api_url("detectors/"), responses.POST, json={"id": "foobar"}, match=match)


def add_mock_detector_edit_response(d_id, **kwargs):
    match = responses.json_params_matcher({"configuration": kwargs}) if kwargs else None
    _add_api_response(detector_api_url("detectors/%s/" % d_id), responses.PUT, status=204, match=match)


def add_mock_detector_train_responses(detector_id):
    _add_api_response(detector_api_url("detectors/%s/train/" % detector_id), responses.POST, OP_RESP)


def add_mock_run_dataset_recommendation_responses(detector_id):
    _add_api_response(
        detector_api_url("detectors/%s/dataset_recommendation/" % detector_id), responses.POST, OP_RESP
    )


def add_mock_operations_responses(status, **kwargs):
    data = {"type": "mock_operation_type", "status": status}
    if kwargs:
        data.update(kwargs)
    if status == "success":
        data.update(
            {
                "metadata": {
                    "raster_id": "foo",
                    "detector_id": "bar",
                    "folder_id": "spam",
                }
            }
        )
    _add_api_response(detector_api_url("operations/%s/" % OPERATION_ID), json=data)


def add_mock_annotations_responses(
    detector_id, raster_id, annotation_type, class_id=None
):
    upload_id = 32
    url = "detectors/%s/training_rasters/%s/%s/upload/bulk/" % (
        detector_id,
        raster_id,
        annotation_type,
    )
    responses.add(responses.PUT, "http://storage.example.com", status=200)
    _add_api_response(
        detector_api_url(url),
        responses.POST,
        {"upload_url": "http://storage.example.com", "upload_id": upload_id},
    )
    url = "detectors/%s/training_rasters/%s/%s/upload/bulk/%s/commit/" % (
        detector_id,
        raster_id,
        annotation_type,
        upload_id,
    )
    if class_id is None:
        _add_api_response(
            detector_api_url(url),
            responses.POST,
            OP_RESP,
            # strict_match matters here because we want to disallow sending `class_id: null`
            # as this would lead to a server-side error. Instead, class_id shouldn't be included
            # if it is not defined
            match=responses.matchers.json_params_matcher({}, strict_match=True),
        )
    else:
        _add_api_response(
            detector_api_url(url),
            responses.POST,
            json=OP_RESP,
            match=responses.matchers.json_params_matcher({"class_id": class_id}),
        )


def add_mock_raster_upload_responses(identity_key, multispectral, cloud_coverage, tag):
    raster_id = 42
    # Upload initiation
    data = {"upload_url": "http://storage.example.com", "raster_id": raster_id}
    body = {
        "name": "test 1",
        "multispectral": multispectral,
        "captured_at": "2020-01-10T12:34:56.789Z",
        "folder_id": "a-folder-uuid",
    }
    if identity_key:
        body["identity_key"] = identity_key
    if cloud_coverage is not None:
        body["cloud_coverage"] = cloud_coverage
    if tag is not None:
        body["user_tag"] = tag
    _add_api_response(
        detector_api_url("rasters/upload/file/"),
        responses.POST,
        data,
        responses.matchers.json_params_matcher(body),
        status=200,
    )
    # Storage PUT
    responses.add(responses.PUT, "http://storage.example.com", status=200)
    # Commit
    _add_api_response(detector_api_url("rasters/%s/commit/" % raster_id), responses.POST, OP_RESP)
    # Status, first check
    data = {"id": raster_id, "name": "raster1", "status": "processing"}
    _add_api_response(detector_api_url("rasters/%s/" % raster_id), json=data)
    # Status, second check
    data = {"id": raster_id, "name": "raster1", "status": "ready"}
    _add_api_response(detector_api_url("rasters/%s/" % raster_id), json=data)


def add_mock_detection_areas_upload_responses(raster_id):
    upload_id = 42

    # Upload initiation
    data = {"upload_url": "http://storage.example.com", "upload_id": upload_id}
    _add_api_response(
        detector_api_url("rasters/%s/detection_areas/upload/file/" % raster_id), responses.POST, data
    )
    # Storage PUT
    responses.add(responses.PUT, "http://storage.example.com", status=200)
    # Commit
    _add_api_response(
        detector_api_url("rasters/%s/detection_areas/upload/%s/commit/" % (raster_id, upload_id)),
        responses.POST,
        OP_RESP,
        status=200,
    )
    # Status, first check
    data = {"status": "processing"}
    _add_api_response(
        detector_api_url("rasters/%s/detection_areas/upload/%s/" % (raster_id, upload_id)), json=data
    )
    # Status, second check
    data = {"status": "ready"}
    _add_api_response(
        detector_api_url("rasters/%s/detection_areas/upload/%s/" % (raster_id, upload_id)), json=data
    )


def add_mock_remote_import_responses(upload_id, post_body):
    match = responses.json_params_matcher(post_body)
    # Upload initiation
    data = {"upload_url": "http://storage.example.com", "upload_id": upload_id}
    _add_api_response(detector_api_url("rasters/import/"), responses.POST, data)
    # Storage PUT
    responses.add(responses.PUT, "http://storage.example.com", status=200)
    # Commit
    _add_api_response(
        detector_api_url(f"rasters/import/{upload_id}/commit/"),
        responses.POST,
        OP_RESP,
        match=match,
        status=200,
    )


def add_mock_detector_run_responses(detector_id, raster_id, secondary_raster_id=None):
    data = {"raster_id": raster_id}
    if secondary_raster_id:
        data["secondary_raster_id"] = secondary_raster_id
    _add_api_response(
        detector_api_url("detectors/%s/run/" % detector_id),
        responses.POST,
        OP_RESP,
        match=responses.matchers.json_params_matcher(data),
    )
    # First status check
    data = {"status": "running"}
    _add_api_response(detector_api_url("operations/%s/" % OPERATION_ID), json=data)
    # Second status check
    data = {"status": "success"}
    _add_api_response(detector_api_url("operations/%s/" % OPERATION_ID), json=data)


def add_mock_vector_layer_responses(upload_id, raster_id, name, color):
    _add_api_response(
        detector_api_url("vector_layers/%s/upload/" % raster_id),
        responses.POST,
        json={"upload_url": "http://storage.example.com", "upload_id": upload_id},
    )
    responses.add(responses.PUT, "http://storage.example.com", status=200)
    qs = {}
    if name is not None:
        qs["name"] = name
    if color is not None:
        qs["color"] = color
    _add_api_response(
        detector_api_url("vector_layers/%s/upload/%s/commit/" % (raster_id, upload_id)),
        responses.POST,
        json={"operation_id": OPERATION_ID, "poll_interval": TEST_POLL_INTERVAL},
        match=responses.matchers.json_params_matcher(qs) if len(qs.keys()) != 0 else [],
    )


def add_mock_vector_layer_download_responses(layer_id, polygons_num):
    url = "vector_layers/%s/download/" % layer_id
    data = {"operation_id": OPERATION_ID, "poll_interval": TEST_POLL_INTERVAL}
    _add_api_response(detector_api_url(url), verb=responses.POST, json=data)
    results = {
        "expiration": "2021-11-03T10:55:16.000000Z",
        "download_url": "http://layer.geojson.example.com",
    }
    add_mock_operations_responses("success", results=results)
    url = results["download_url"]
    polygons_fc = multipolygon_to_polygon_feature_collection(make_geojson_multipolygon(polygons_num))
    assert len(polygons_fc["features"]) == polygons_num
    responses.add(
        responses.GET,
        url,
        body=json.dumps(polygons_fc),
    )
    return polygons_fc

def add_mock_folder_creation_response(id, name):
    match = responses.matchers.json_params_matcher({"name": name})
    _add_api_response(detector_api_url("folders/"), responses.POST, json={"id": id}, match=match)


def make_geojson_polygon(base=1):
    return {"type": "Polygon", "coordinates": [[[0, 0], [base, 0], [base, base], [0, base], [0, 0]]]}


def make_geojson_multipolygon(npolygons=1):
    coords = []
    for i in range(npolygons):
        coords.append(make_geojson_polygon(i + 1)["coordinates"])
    return {"type": "MultiPolygon", "coordinates": coords}


def add_mock_download_result_response(op_id, num_classes):
    data = {
        "results": {
            "url": "http://storage.example.com/result_for_class_1.geojson",
            "by_class": [
                {
                    "class": {"name": f"class_{i + 1}"},
                    "result": {
                        "url": f"http://storage.example.com/result_for_class_{i + 1}.geojson",
                        "vector_layer_id": f"layer_{i + 1}"
                    },
                } for i in range(num_classes)
            ],
        },
    }
    _add_api_response(detector_api_url("operations/%s/" % op_id), json=data, status=201)
    mock_contents = {
        f"class_{i + 1}": json.dumps(make_geojson_multipolygon(npolygons=i + 2))
        for i in range(num_classes)
    }
    for i in range(num_classes):
        responses.add(
            responses.GET,
            f"http://storage.example.com/result_for_class_{i + 1}.geojson",
            body=mock_contents[f"class_{i + 1}"],
        )
    return mock_contents


def add_mock_download_raster_response(raster_id):
    file_url = "http://storage.example.com/%s.tiff" % raster_id
    data = {"download_url": file_url}
    _add_api_response(detector_api_url("rasters/%s/download/" % raster_id), json=data)
    mock_content = (1024).to_bytes(2, byteorder="big")
    responses.add(responses.GET, file_url, body=mock_content)
    return mock_content


def add_mock_url_result_response(op_id, url):
    data = {"results": {"url": url}}
    _add_api_response(detector_api_url("operations/%s/" % op_id), json=data, status=201)


def add_get_operation_results_url_response(op_id):
    url = "http://storage.example.com/42.geojson"
    data = {"results": {"url": url}}
    _add_api_response(detector_api_url("operations/%s/" % op_id), json=data, status=201)
    return url


def add_mock_edit_raster_response(raster_id, body):
    _add_api_response(
        detector_api_url("rasters/%s/" % raster_id),
        responses.PUT,
        match=responses.matchers.json_params_matcher(body),
        status=204,
    )


def add_mock_delete_raster_response(raster_id):
    _add_api_response(detector_api_url("rasters/%s/" % raster_id), responses.DELETE)


def add_mock_delete_detectionarea_response(raster_id):
    _add_api_response(detector_api_url("rasters/%s/detection_areas/" % raster_id), responses.DELETE)


def add_mock_delete_detector_response(detector_id):
    _add_api_response(detector_api_url("detectors/%s/" % detector_id), responses.DELETE)


def add_mock_delete_vector_layer_response(layer_id):
    _add_api_response(detector_api_url("vector_layers/%s/" % layer_id), responses.DELETE)


def add_mock_edit_vector_layer_response(layer_id, **kwargs):
    _add_api_response(
        detector_api_url("vector_layers/%s/" % layer_id),
        responses.PUT,
        match=responses.matchers.json_params_matcher(kwargs),
    )


def add_mock_raster_markers_list_response(raster_id):
    base_url = detector_api_url("rasters/%s/markers/" % raster_id)
    data1 = {
        "count": 4,
        "next": base_url + "?page_number=2",
        "previous": base_url + "?page_number=1",
        "page_size": 2,
        "results": [{"id": "1"}, {"id": "2"}],
    }
    data2 = {
        "count": 4,
        "next": None,
        "previous": None,
        "page_size": 2,
        "results": [{"id": "3"}, {"id": "4"}],
    }
    _add_api_response(
        base_url,
        json=data1,
        match=responses.matchers.query_param_matcher({"page_number": "1"}),
    )
    _add_api_response(
        base_url,
        json=data2,
        match=responses.matchers.query_param_matcher({"page_number": "2"}),
    )


def add_mock_marker_creation_response(marker_id, raster_id, detector_id, coords, text):
    if detector_id is None:
        url = "rasters/%s/markers/" % raster_id
    else:
        url = "detectors/%s/training_rasters/%s/markers/" % (detector_id, raster_id)
    body = {
        "marker": {"type": "Point", "coordinates": coords},
        "text": text,
    }
    match = responses.matchers.json_params_matcher(body)
    _add_api_response(detector_api_url(url), responses.POST, json={"id": marker_id}, match=match)


def add_mock_folder_detector_response(folder_id: str):
    base_url = detector_api_url("folders/%s/detectors/" % folder_id)
    data1 = {
        "count": 4,
        "next": base_url + "?page_number=2",
        "previous": base_url + "?page_number=1",
        "page_size": 2,
        "results": [
            {
                "id": "id1",
                "name": "detector1",
                "is_runnable": True,
                "user_tag": "tag1",
            },
            {
                "id": "id2",
                "name": "detector2",
                "is_runnable": False,
                "user_tag": "tag2",
            },
        ],
    }
    data2 = {
        "count": 4,
        "next": None,
        "previous": None,
        "page_size": 2,
        "results": [
            {
                "id": "id3",
                "name": "detector3",
                "is_runnable": True,
                "user_tag": "",
            },
            {
                "id": "id4",
                "name": "detector4",
                "is_runnable": False,
                "user_tag": "",
            },
        ],
    }
    _add_api_response(
        base_url,
        json=data1,
        match=responses.matchers.query_param_matcher({"page_number": "1"}),
    )
    _add_api_response(
        base_url,
        json=data2,
        match=responses.matchers.query_param_matcher({"page_number": "2"}),
    )


def test_multipolygon_to_polygon_feature_collection():
    mp = {
        "type": "MultiPolygon",
        "coordinates": [
            [[[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]]],
            [[[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]]]
        ]
    }
    fc = multipolygon_to_polygon_feature_collection(mp)
    assert fc == {
        "type": "FeatureCollection",
        "features": [{
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[0, 0], [0, 1], [1, 1], [1, 0], [0, 0]]]
            }
        },  {
            "type": "Feature",
            "properties": {},
            "geometry": {
                "type": "Polygon",
                "coordinates": [[[1, 1], [1, 2], [2, 2], [2, 1], [1, 1]]]
            }
        }]
    }


@pytest.mark.parametrize(
    ("identity_key", "multispectral", "cloud_coverage", "tag"),
    ((None, False, None, None), ("abc", True, 18, "spam")),
)
@responses.activate
def test_upload_raster(monkeypatch, identity_key, multispectral, cloud_coverage, tag):
    client = _client(monkeypatch)
    add_mock_raster_upload_responses(identity_key, multispectral, cloud_coverage, tag)
    add_mock_operations_responses("success")
    with tempfile.NamedTemporaryFile() as f:
        # This just tests that this doesn't raise
        client.upload_raster(
            f.name,
            name="test 1",
            folder_id="a-folder-uuid",
            captured_at="2020-01-10T12:34:56.789Z",
            identity_key=identity_key,
            multispectral=multispectral,
            cloud_coverage=cloud_coverage,
            user_tag=tag,
        )
    assert len(responses.calls) == 4


@pytest.mark.parametrize(
    "edited_data",
    (
        {"folder_id": "2233"},
        {"folder_id": "2233", "identity_key": "dr43t5zrtzz"},
        {"captured_at": "2020-01-01T12:34:56.789Z", "cloud_coverage": 88},
        {
            "multispectral_band_specification": {
                "ranges": [[2, 3], [12, 13], [22, 23]],
                "vizbands": [0, 1, 2],
            }
        },
        {"user_tag": "foobar"},
    ),
)
@responses.activate
def test_edit_raster(monkeypatch, edited_data):
    RASTER_ID = "foobar"
    client = _client(monkeypatch)
    add_mock_edit_raster_response(RASTER_ID, {"name": "spam", **edited_data})
    client.edit_raster(RASTER_ID, "spam", **edited_data)
    assert len(responses.calls) == 1


@responses.activate
def test_get_raster(monkeypatch):
    """Test the raster information"""
    RASTER_ID = "foobar"
    client = _client(monkeypatch)
    _add_api_response(detector_api_url("rasters/%s/" % RASTER_ID), json={}, status=201)
    client.get_raster(RASTER_ID)
    assert len(responses.calls) == 1


@responses.activate
def test_delete_raster(monkeypatch):
    RASTER_ID = "foobar"
    client = _client(monkeypatch)
    add_mock_delete_raster_response(RASTER_ID)
    client.delete_raster(RASTER_ID)
    assert len(responses.calls) == 1


@responses.activate
def test_delete_detectionarea(monkeypatch):
    RASTER_ID = "foobar"
    client = _client(monkeypatch)
    add_mock_delete_detectionarea_response(RASTER_ID)
    client.remove_raster_detection_areas(RASTER_ID)
    assert len(responses.calls) == 1


@responses.activate
def test_download_raster(monkeypatch):
    RASTER_ID = "foobar"
    expected_content = add_mock_download_raster_response(RASTER_ID)
    client = _client(monkeypatch)
    with tempfile.NamedTemporaryFile() as f:
        client.download_raster_to_file(RASTER_ID, f.name)
        assert open(f.name, "rb").read() == expected_content
    assert len(responses.calls) == 2


@responses.activate
def test_list_rasters(monkeypatch):
    """Test the list of rasters, both generic and specifying the filters"""
    client = _client(monkeypatch)
    # Generic (check pagination)
    add_mock_rasters_list_response()
    page1 = client.list_rasters()
    assert len(page1) == 2
    assert page1[0]["name"] == "raster1" and page1[1]["name"] == "raster2"
    assert len(responses.calls) == 1
    page2 = page1.next()
    assert len(page2) == 2
    assert page2[0]["name"] == "raster3" and page2[1]["name"] == "raster4"
    page3 = page2.next()
    assert len(responses.calls) == 3
    assert len(page3) == 1 and page3.next() is None
    assert page3[0]["name"] == "raster5"
    assert len(responses.calls) == 3
    pairs = zip(list(page2), list(client.list_rasters(page_number=2)))
    assert all(x == y for x, y in pairs) is True
    assert len(responses.calls) == 4
    # Folder list
    add_mock_rasters_in_folder_list_response("foobar")
    rasters = client.list_rasters("foobar")
    assert rasters[0]["name"] == "raster_in_folder1"
    assert rasters[0]["folder_id"] == "foobar"
    # Search list
    add_mock_rasters_in_filtered_list_response(search="spam")
    rasters = client.list_rasters("", search_string="spam")
    assert rasters[0]["name"] == "spam_raster"
    # Filter list 1
    add_mock_rasters_in_filtered_list_response(tag="foobar")
    rasters = client.list_rasters("", user_tag="foobar")
    assert rasters[0]["name"] == "raster_foobar"
    # Filter list 2
    add_mock_rasters_in_filtered_list_response(tag="foobar", cloud=44)
    rasters = client.list_rasters("", user_tag="foobar", max_cloud_coverage=44)
    assert rasters[0]["name"] == "raster_foobar"
    # Filter list 3
    add_mock_rasters_in_filtered_list_response(
        has_layers=False, search="foo", before="2018-11-13T20:20:39+00:00"
    )
    rasters = client.list_rasters(
        search_string="foo",
        captured_before="2018-11-13T20:20:39+00:00",
        has_vector_layers=False,
    )
    assert rasters[0]["name"] == "foo_raster"
    # # Filter list 4
    add_mock_rasters_in_filtered_list_response(
        has_layers=True, after="2022-11-13T20:20:39+00:00", search="bar"
    )
    rasters = client.list_rasters(
        search_string="bar",
        captured_after="2022-11-13T20:20:39+00:00",
        has_vector_layers=True,
    )
    assert rasters[0]["name"] == "bar_raster"
    # Filter list with pagination
    add_mock_rasters_in_filtered_list_response(page=3, search="spam")
    rasters = client.list_rasters(
        search_string="spam",
        page_number=3,
    )
    assert rasters[0]["id"] == "77"
    assert len(responses.calls) == 11


@responses.activate
def test_detector_creation(monkeypatch):
    client = _client(monkeypatch)
    args = [
        {"detection_type": "segmentation"},
        {"output_type": "bbox"},
        {"training_steps": 10**3},
        {"backbone": "resnet18"},
        {"tile_size": 352},
        {"background_sample_ratio": 0.3},
    ]
    add_mock_detector_creation_response()
    client.create_detector()
    for a in args:
        add_mock_detector_creation_response(**a)
        client.create_detector(**a)
    merge = dict(p for d in args for p in d.items())
    add_mock_detector_creation_response(**merge)
    detector_id = client.create_detector(**merge)
    assert detector_id == "foobar"


@responses.activate
def test_list_detectors(monkeypatch):
    client = _client(monkeypatch)
    # Full list
    add_mock_detectors_list_response()
    detectors = client.list_detectors()
    assert len(detectors) == 2  # 1st api call
    assert detectors[0]["name"] == "detector1"
    assert detectors[1]["name"] == "detector2"
    assert detectors.next()[1]["id"] == "43"  # 2nd api call
    # Search list
    add_mock_detectors_list_response("spam")
    detectors = client.list_detectors("spam")  # 3rd api call
    assert detectors[0]["name"] == "spam"
    # Filter list
    add_mock_detectors_list_response(None, "foobar", True)
    detectors = client.list_detectors(user_tag="foobar", is_shared=True)  # 4th api call
    assert detectors[1]["name"] == "detector2"
    assert len(responses.calls) == 4


@responses.activate
def test_delete_detector(monkeypatch):
    DETECTOR_ID = "foobar"
    client = _client(monkeypatch)
    add_mock_delete_detector_response(DETECTOR_ID)
    client.delete_detector(DETECTOR_ID)
    assert len(responses.calls) == 1


@responses.activate
def test_detector_edit(monkeypatch):
    client = _client(monkeypatch)
    detector_id = "foobar"
    args = [
        {"detection_type": "segmentation"},
        {"output_type": "bbox"},
        {"training_steps": 10**3},
        {"backbone": "resnet50"},
        {"tile_size": 512},
        {"background_sample_ratio": 0.3},
    ]
    add_mock_detector_edit_response(detector_id)
    client.edit_detector(detector_id)
    for a in args:
        add_mock_detector_edit_response(detector_id, **a)
        client.edit_detector(detector_id, **a)
    merge = dict(p for d in args for p in d.items())
    add_mock_detector_edit_response(detector_id, **merge)
    client.edit_detector(detector_id, **merge)
    assert len(responses.calls) == 8


@responses.activate
def test_set_raster_detection_areas_from_file(monkeypatch):
    add_mock_detection_areas_upload_responses(1)
    add_mock_operations_responses("success")

    client = _client(monkeypatch)
    # This just tests that this doesn't raise
    with tempfile.NamedTemporaryFile() as f:
        client.set_raster_detection_areas_from_file(1, f.name)
    assert len(responses.calls) == 4


@responses.activate
def test_run_detector(monkeypatch):
    add_mock_detector_run_responses(1, 2)
    client = _client(monkeypatch)
    client.run_detector(1, 2)
    assert len(responses.calls) == 3


@responses.activate
def test_run_detector_secondary_raster(monkeypatch):
    add_mock_detector_run_responses(1, 2, 3)
    client = _client(monkeypatch)
    client.run_detector(1, 2, 3)
    assert len(responses.calls) == 3


@responses.activate
def test_download_result_to_file(monkeypatch):
    expected_content = add_mock_download_result_response(101, 1)["class_1"]
    client = _client(monkeypatch)
    with tempfile.NamedTemporaryFile() as f:
        client.download_result_to_file(101, f.name)
        assert open(f.name).read() == expected_content
    assert len(responses.calls) == 2


@responses.activate
def test_download_result_to_feature_collection(monkeypatch):
    add_mock_download_result_response(101, 2)
    add_mock_vector_layer_download_responses("layer_1", 10)
    add_mock_vector_layer_download_responses("layer_2", 20)
    client = _client(monkeypatch)
    with tempfile.NamedTemporaryFile() as f:
        client.download_result_to_feature_collection(101, f.name)
        with open(f.name) as fr:
            fc = json.load(fr)
        assert fc["type"] == "FeatureCollection" and len(fc["features"]) == 2
        class_1_index = (
            0 if fc["features"][0]["properties"]["class_name"] == "class_1" else 1
        )
        feat1 = fc["features"][class_1_index]
        assert feat1["type"] == "Feature"
        assert feat1["properties"]["class_name"] == "class_1"
        assert feat1["geometry"] == make_geojson_multipolygon(10)
        assert len(feat1["geometry"]["coordinates"]) == 10
        assert isinstance(feat1["geometry"]["coordinates"][0][0][0][0], (int, float))
        feat2 = fc["features"][(class_1_index + 1) % 2]
        assert feat2["type"] == "Feature" and feat2["geometry"]["type"] == "MultiPolygon"
        assert feat2["properties"]["class_name"] == "class_2"
        assert len(feat2["geometry"]["coordinates"]) == 20
        assert isinstance(feat2["geometry"]["coordinates"][0][0][0][0], (int, float))
    assert len(responses.calls) == 7


@responses.activate
@pytest.mark.parametrize(
    "annotation_type", ["outline", "training_area", "testing_area", "validation_area"]
)
def test_upload_annotations(monkeypatch, annotation_type):
    add_mock_annotations_responses(1, 2, annotation_type)
    add_mock_operations_responses("running")
    add_mock_operations_responses("running")
    add_mock_operations_responses("success")
    client = _client(monkeypatch)
    client.set_annotations(1, 2, annotation_type, {})
    assert len(responses.calls) == 6


@responses.activate
def test_upload_annotations_class_id(monkeypatch):
    add_mock_annotations_responses(1, 2, "outline", class_id="42")
    add_mock_operations_responses("success")
    client = _client(monkeypatch)
    client.set_annotations(1, 2, "outline", {}, class_id="42")


@responses.activate
def test_train_detector(monkeypatch):
    add_mock_detector_train_responses(1)
    add_mock_operations_responses("running")
    add_mock_operations_responses("running")
    add_mock_operations_responses(
        "success",
        results={
            "score": 92,
            "stats": {
                "rasters_count": 1,
                "training_areas_count": 2,
                "assessment_areas_count": 10,
                "validation_areas_count": 5,
                "total_annotations_count": 4,
                "training_annotations_count": 3,
                "validation_annotations_count": 1,
            },
        },
    )
    client = _client(monkeypatch)
    op = client.train_detector(1)
    assert op["results"]["score"] == 92
    assert len(responses.calls) == 4


@responses.activate
def test_run_dataset_recommendation(monkeypatch):
    add_mock_run_dataset_recommendation_responses(1)
    add_mock_operations_responses("running")
    add_mock_operations_responses("running")
    add_mock_operations_responses("success")
    client = _client(monkeypatch)
    op = client.run_dataset_recommendation(1)
    assert op["status"] == "success"
    assert len(responses.calls) == 4


@pytest.mark.parametrize(("name", "color"), ((None, None), ("foobar", "#aabbcc")))
@responses.activate
def test_upload_vector_layer(monkeypatch, name, color):
    add_mock_vector_layer_responses(11, 22, name, color)
    add_mock_operations_responses("running")
    add_mock_operations_responses("success", results={"vector_layer_id": "spam"})
    client = _client(monkeypatch)
    with tempfile.NamedTemporaryFile() as f:
        assert client.upload_vector_layer(22, f.name, name, color) == "spam"
    assert len(responses.calls) == 5  # upload req, upload PUT, commit + 2 op polling


@responses.activate
def test_delete_vector_layer(monkeypatch):
    LAYER_ID = "foobar"
    client = _client(monkeypatch)
    add_mock_delete_vector_layer_response(LAYER_ID)
    client.delete_vector_layer(LAYER_ID)
    assert len(responses.calls) == 1


@responses.activate
def test_edit_vector_layer(monkeypatch):
    LAYER_ID = "foobar"
    client = _client(monkeypatch)
    add_mock_edit_vector_layer_response(LAYER_ID, color="#ffffff", name="spam")
    client.edit_vector_layer(LAYER_ID, color="#ffffff", name="spam")
    assert len(responses.calls) == 1


@responses.activate
def test_download_vector_layer_to_file(monkeypatch):
    polygons_fc = add_mock_vector_layer_download_responses("foobar", 2)
    client = _client(monkeypatch)
    with tempfile.NamedTemporaryFile() as fp:
        client.download_vector_layer_to_file("foobar", fp.name)
        fc = json.load(fp)
    assert fc["type"] == "FeatureCollection"
    assert fc == polygons_fc and len(fc["features"]) == 2
    assert fc["features"][0]["geometry"]["type"] == "Polygon"
    assert isinstance(fc["features"][1]["geometry"]["coordinates"][0][0][0], (int, float))
    assert len(responses.calls) == 3  # POST /download, GET /operations, GET url


@responses.activate
def test_list_raster_markers(monkeypatch):
    client = _client(monkeypatch)
    add_mock_raster_markers_list_response("spam")
    rasters = client.list_raster_markers("spam")
    assert rasters[0]["id"] == "1"
    rasters = client.list_raster_markers("spam", page_number=2)
    assert rasters[0]["id"] == "3"
    assert len(responses.calls) == 2


@responses.activate
def test_raster_markers_creation(monkeypatch):
    client = _client(monkeypatch)
    add_mock_marker_creation_response("spam", "foo", "bar", [12.34, 56.78], "foobar")
    marker = client.create_marker("foo", "bar", 12.34, 56.78, "foobar")
    assert marker["id"] == "spam"


@responses.activate
def test_create_raster_marker(monkeypatch):
    client = _client(monkeypatch)
    add_mock_marker_creation_response(
        "id123", "rasterid123", None, [43.21, 87.65], "comment"
    )
    marker = client.create_marker("rasterid123", None, 43.21, 87.65, "comment")
    assert marker["id"] == "id123"


@responses.activate
def test_list_folder_detectors(monkeypatch):
    client = _client(monkeypatch)
    add_mock_folder_detector_response("folder_id123")
    detector_list = client.list_folder_detectors("folder_id123")
    assert len(detector_list) == 2
    assert detector_list[0]["id"] == "id1"
    detector_list = detector_list.next()
    assert detector_list[0]["id"] == "id3"
    assert len(responses.calls) == 2


@responses.activate
def test_list_raster_vector_layers(monkeypatch):
    client = _client(monkeypatch)
    add_mock_vector_layers_filtered_list_response(0, "raster1")
    add_mock_vector_layers_filtered_list_response(1, "raster1", "spam", "detector1")
    assert client.list_raster_vector_layers("raster1")[0]["id"] == "0"
    assert (
        client.list_raster_vector_layers("raster1", "spam", "detector1")[0]["name"]
        == "layer_1"
    )


@responses.activate
def test_run_advanced_tool(monkeypatch):
    _add_api_response(
        detector_api_url("advanced_tools/foobar/run/"),
        responses.POST,
        json=OP_RESP,
        match=responses.matchers.json_params_matcher(
            {
                "inputs": {"foo": "bar"},
                "outputs": {"spam": [1, 2], "bar": {"foo": None, "bar": 4}},
            }
        ),
    )
    add_mock_operations_responses("success")
    client = _client(monkeypatch)
    assert (
        client.run_advanced_tool(
            "foobar", {"foo": "bar"}, {"spam": [1, 2], "bar": {"foo": None, "bar": 4}}
        )["type"]
        == "mock_operation_type"
    )
    assert len(responses.calls) == 2


@responses.activate
def test_import_raster_from_remote_source(monkeypatch):
    body = {
        "method": "streaming",
        "source_id": "source",
        "folder_id": "project",
        "name": "image",
    }
    add_mock_remote_import_responses("upload_id", body)
    add_mock_operations_responses("success")

    client = _client(monkeypatch)
    # This just tests that this doesn't raise
    with tempfile.NamedTemporaryFile() as f:
        assert (
            client.import_raster_from_remote_source(
                "image", "project", "source", f.name
            )
            == "foo"
        )
    assert len(responses.calls) == 4


@responses.activate
def test_list_detector_rasters(monkeypatch):
    client = _client(monkeypatch)
    add_mock_rasters_list_response(detector_api_url("detectors/spam/training_rasters/"))
    page1 = client.list_detector_rasters("spam")
    assert page1[0]["name"] == "raster1" and page1[1]["name"] == "raster2"
    page2 = client.list_detector_rasters("spam", page_number=2)
    assert page2[0]["name"] == "raster3" and page2[1]["name"] == "raster4"
    assert len(responses.calls) == 2


@responses.activate
def test_folder_creation(monkeypatch):
    client = _client(monkeypatch)
    add_mock_folder_creation_response("folder-id", "folder-name")
    assert client.create_folder("folder-name") == "folder-id"



"""
Handles interfacing with the API documented at https://app.picterra.ch/public/apidocs/plots_analysis/v1/

Note that Tracer is separate from Forge and so an API key which is valid for
one may encounter permissions issues if used with the other
"""
import datetime
import json
import os.path
import sys
import warnings

if sys.version_info >= (3, 8):
    from typing import Any, Dict, List, Literal, Optional, Tuple
else:
    from typing_extensions import Literal
    from typing import Any, Dict, List, Optional, Tuple

import requests

from picterra.base_client import (
    BaseAPIClient,
    ResultsPage,
    _check_resp_is_ok,
    _download_to_file,
)


class TracerClient(BaseAPIClient):
    def __init__(self, **kwargs):
        super().__init__("public/api/plots_analysis/v1/", **kwargs)

    def _return_results_page(
        self, resource_endpoint: str, params: Optional[Dict[str, Any]] = None
    ) -> ResultsPage:
        if params is None:
            params = {}
        if "page_number" not in params:
            params["page_number"] = 1

        url = self._full_url("%s/" % resource_endpoint, params=params)
        return ResultsPage(url, self.sess.get)

    def _make_upload(self) -> Tuple[str, str]:
        resp = self.sess.post(self._full_url("upload/file/"))
        _check_resp_is_ok(resp, "Failure obtaining an upload")
        upload_id, upload_url = resp.json()["upload_id"], resp.json()["upload_url"]
        return upload_id, upload_url

    def list_methodologies(
        self,
        search: Optional[str] = None,
        page_number: Optional[int] = None,
    ) -> ResultsPage:
        """
        List all the methodologies the user can access, see `ResultsPage`
            for the pagination access pattern.

        Args:
            search: The term used to filter methodologies by name
            page_number: Optional page (from 1) of the list we want to retrieve

        Returns:
            See https://app.picterra.ch/public/apidocs/plots_analysis/v1/#tag/plots-groups/operation/getMethodologiesList

        Example:

            ::

                {
                    'id': '42',
                    'name': 'Coffee - EUDR',
                },
                {
                    'id': '43',
                    'name': 'Cattle - EUDR'
                }

        """
        data: Dict[str, Any] = {}
        if search is not None:
            data["search"] = search.strip()
        if page_number is not None:
            data["page_number"] = int(page_number)
        return self._return_results_page("methodologies", data)

    def create_plots_group(
        self,
        plots_group_name: str,
        methodology_id: str,
        plots_geometries_filenames: List[str],
        columns: Optional[Dict[str, str]] = None
    ) -> str:
        """
        Creates a new plots group.

        Args:
            plots_group_name: user-friendly name for the group
            methodology_id: id of the methodology to use, retrieved via list_methodologies
            plots_geometries_filenames: Paths to files containing the geometries of the plots the group will have
            columns: columns to add to the group. if any

        Returns:
            str: the id of the new group.
        """
        data = {
            "name": plots_group_name,
            "methodology_id": methodology_id,
            "custom_columns_values": columns or {}
        }
        resp = self.sess.post(self._full_url("plots_groups/"), json=data)
        _check_resp_is_ok(resp, "Failure starting plots group commit")
        op_result = self._wait_until_operation_completes(resp.json())["results"]
        self.update_plots_group_plots(op_result["plots_group_id"], plots_geometries_filenames)
        return op_result["plots_group_id"]

    def update_plots_group_plots(self, plots_group_id: str, plots_geometries_filenames: List[str], delete_existing_plots: bool = False):
        """
        Updates the geometries of a given plots group

        Args:
            plots_group_id: identifier for the plots group to replace
            plots_geometries_filenames: List of paths to files containing the geometries of the plots the group will have
            delete_existing_plots: If true, will remove all existing plots in the plots group before uploading new ones.
                If False (default), plot data uploaded is merged with existing plots.

        Returns:
            dict: The creation operation result, which includes the plot group id
        """
        files = []
        for filename in plots_geometries_filenames:
            upload_id, upload_url = self._make_upload()
            with open(filename, "rb") as fh:
                resp = requests.put(upload_url, data=fh.read())
                _check_resp_is_ok(resp, "Failure uploading plots file for group")
                files.append({"filename": os.path.basename(filename), "upload_id": upload_id})
        data = {"files": files, "overwrite": delete_existing_plots}
        resp = self.sess.post(self._full_url(f"plots_groups/{plots_group_id}/upload/commit/"), json=data)
        _check_resp_is_ok(resp, "Failure starting plots group update:")
        return self._wait_until_operation_completes(resp.json())

    def download_plots_group_to_file(self, plots_group_id: str, format: Literal["excel", "geojson"], filename: str) -> None:
        """
        Downloads a plots group to a local file

        Args:
            plots_group_id: The id of the plots group to download
            filename: The local filename where to save the plots group

        Raises:
            APIError: There was an error while trying to download the plots group id
        """
        data = {"format": format}
        resp = self.sess.post(self._full_url("plots_groups/%s/export/" % plots_group_id), json=data)
        _check_resp_is_ok(resp, "Failure starting plots group download")
        op = self._wait_until_operation_completes(resp.json())
        _download_to_file(op["results"]["download_url"], filename)

    def list_plots_groups(
        self,
        search: Optional[str] = None,
        page_number: Optional[int] = None,
        include_archived: bool = False,
    ) -> ResultsPage:
        """
        List all the plots group the user can access, see `ResultsPage`
            for the pagination access pattern.

        This function is still **beta** and subject to change.


        Args:
            search: The term used to filter by name
            page_number: Optional page (from 1) of the list we want to retrieve
            include_archived: If true, includes archived plot groups in the results

        Returns:
            See https://app.picterra.ch/public/apidocs/plots_analysis/v1/#tag/plots-groups/operation/getPlotsGroupsList
        """
        data: Dict[str, Any] = {}
        if search is not None:
            data["search"] = search.strip()
        if page_number is not None:
            data["page_number"] = int(page_number)
        if include_archived:
            data["include_archived"] = include_archived
        return self._return_results_page("plots_groups", data)

    def analyze_plots_precheck(
        self,
        plots_group_id: str,
        plots_analysis_name: str,
        plot_ids: List[str],
        date_from: datetime.date,
        date_to: datetime.date
    ) -> dict:
        """
        Check the analysis for a given date over the plot ids of the specified plot group has no errors

        Args:
            plots_group_id: id of the plots group on which we want to run the new analysis
            plots_analysis_name: name to give to the new analysis
            plot_ids: list of the plot ids of the plots group to select for the analysis
            date_from: start point in time at which the analysis should be evaluated; please note that
                **the date that make sense are methodology dependent**, so please check the methodology
                of the plots group beforehand
            date_to: end point in time at which the analysis should be evaluated.

        Returns:
            dict: the precheck data
        """
        upload_id, upload_url = self._make_upload()
        resp = requests.put(upload_url, data=json.dumps({"plot_ids": plot_ids}))
        _check_resp_is_ok(resp, "Failure uploading plots file for analysis")
        data = {
            "analysis_name": plots_analysis_name,
            "upload_id": upload_id,
            "date_from": date_from.isoformat(),
            "date_to": date_to.isoformat()
        }
        resp = self.sess.post(self._full_url(f"plots_groups/{plots_group_id}/analysis/precheck/"), json=data)
        _check_resp_is_ok(resp, "Failure starting analysis precheck")
        op_result = self._wait_until_operation_completes(resp.json())
        url = op_result["results"]["precheck_data_url"]
        return requests.get(url).json()

    def analyze_plots(
        self,
        plots_group_id: str,
        plots_analysis_name: str,
        plot_ids: List[str],
        date_from: datetime.date,
        date_to: datetime.date
    ) -> str:
        """
        Runs the analysis for a given date over the plot ids of the specified plot group,
        and returns the URL where we can see the analysis in the Picterra platform.

        Args:
            plots_group_id: id of the plots group on which we want to run the new analysis
            plots_analysis_name: name to give to the new analysis
            plot_ids: list of the plot ids of the plots group to select for the analysis
            date_from: start point in time at which the analysis should be evaluated; please note
                that **the date that make sense are methodology dependent**, so please check the
                methodology of the plots group beforehand
            date_to: end point in time at which the analysis should be evaluated.

        Returns:
            dict: the analysis id.
        """
        upload_id, upload_url = self._make_upload()
        resp = requests.put(upload_url, data=json.dumps({"plot_ids": plot_ids}))
        _check_resp_is_ok(resp, "Failure uploading plots file for analysis")
        data = {
            "analysis_name": plots_analysis_name,
            "upload_id": upload_id,
            "date_from": date_from.isoformat(),
            "date_to": date_to.isoformat()
        }
        resp = self.sess.post(self._full_url(f"plots_groups/{plots_group_id}/analysis/"), json=data)
        _check_resp_is_ok(resp, "Couldn't start analysis")
        op_result = self._wait_until_operation_completes(resp.json())
        analysis_id = op_result["results"]["analysis_id"]
        return analysis_id

    def list_plots_analyses(
        self,
        plots_group_id: str,
        search: Optional[str] = None,
        page_number: Optional[int] = None,
        include_archived: bool = False,
    ) -> ResultsPage:
        """
        List all the plots analyses the user can access, see `ResultsPage`
            for the pagination access pattern.

        This function is still **beta** and subject to change.

        Args:
            plots_group_id: id of the plots group on which we want to list the analyses
            search: The term used to filter by name
            page_number: Optional page (from 1) of the list we want to retrieve
            include_archived: Defaults to false. If true, includes archived analyses in the results

        Returns:
            See https://app.picterra.ch/public/apidocs/plots_analysis/v1/#tag/analysis/operation/getPlotsAnalysesList
        """
        data: Dict[str, Any] = {}
        if include_archived:
            data["include_archived"] = str(include_archived).lower()
        if search is not None:
            data["search"] = search.strip()
        if page_number is not None:
            data["page_number"] = int(page_number)
        return self._return_results_page(f"plots_groups/{plots_group_id}/analysis/", data)

    def list_plots_analysis_reports(
        self,
        plots_analysis_id: str,
        plots_group_id: Optional[str] = None,
        page_number: Optional[int] = None,
        include_archived: bool = False,
    ) -> ResultsPage:
        """
        List all the reports belonging to a given plots analysis, see `ResultsPage`
            for the pagination access pattern.

        Args:
            plots_analysis_id: id of the plots analysis for which we want to list the reports
            page_number: Optional page (from 1) of the list we want to retrieve
            include_archived: Defaults to false. If true, includes archived analysis reports in the
                              results

        Deprecated arguments:
            plots_group_id: ignored, do not provide it

        Returns:
            See https://app.picterra.ch/public/apidocs/plots_analysis/v1/#tag/reports/operation/getReportsList
        """  # noqa[E501]
        if plots_group_id is not None:
            warnings.warn("Passing plots_group_id is not needed anymore, remove it", DeprecationWarning)

        params: Dict[str, Any] = {}
        if page_number is not None:
            params["page_number"] = int(page_number)
        if include_archived:
            params["include_archived"] = include_archived
        return self._return_results_page(
            f"plots_analyses/{plots_analysis_id}/reports/", params
        )

    def list_plots_analysis_report_types(
        self,
        plots_analysis_id: str,
        plots_group_id: Optional[str] = None,
    ) -> List[Dict[str, Any]]:
        """
        List all the plots analyses report types the user can use (see create_plots_analysis_report)

        Args:
            plots_analysis_id: id of the plots analysis

        Deprecated arguments:
            plots_group_id: ignored, do not provide it

        Returns:
            See https://app.picterra.ch/public/apidocs/plots_analysis/v1/#tag/reports/operation/getReportTypesForAnalysis
        """  # noqa[E501]
        if plots_group_id is not None:
            warnings.warn("Passing plots_group_id is not needed anymore, remove it", DeprecationWarning)

        resp = self.sess.get(
            self._full_url(f"plots_analyses/{plots_analysis_id}/reports/types/")
        )
        _check_resp_is_ok(resp, "Couldn't list report types")
        return resp.json()

    def create_plots_analysis_report_precheck(
        self,
        plots_analysis_id: str,
        report_name: str,
        plot_ids: List[str],
        report_type: str,
        plots_group_id: Optional[str] = None,
        *,
        metadata: Optional[dict] = None
    ) -> Dict[str, Any]:
        """
        Check creation of a report with the given parameters is ok

        If the function fails, the report is not valid

        Args:
            plots_analysis_id: id of the plots analysis
            report_name: name to give to the report
            plot_ids: list of the plot ids to select for the report
            report_type: type of report to generate, as per list_plots_analyses_report_types
            metadata:  set of key-value pairs which may be included in the report

        Deprecated arguments:
            plots_group_id: ignored, do not provide it

        Returns:
            dict: the precheck data
        """
        if plots_group_id is not None:
            warnings.warn("Passing plots_group_id is not needed anymore, remove it", DeprecationWarning)

        upload_id, upload_url = self._make_upload()
        resp = requests.put(upload_url, data=json.dumps({"plot_ids": plot_ids}))
        _check_resp_is_ok(resp, "Failure uploading plots file for analysis")
        data = {
            "name": report_name,
            "upload_id": upload_id,
            "report_type": report_type,
            "metadata": metadata if metadata is not None else {}
        }
        resp = self.sess.post(
            self._full_url(f"plots_analyses/{plots_analysis_id}/reports/precheck/"),
            json=data
        )
        _check_resp_is_ok(resp, "Failure starting precheck")
        self._wait_until_operation_completes(resp.json())
        return {"status": "passed"}

    def create_plots_analysis_report(
        self,
        plots_analysis_id: str,
        report_name: str,
        plot_ids: List[str],
        report_type: str,
        plots_group_id: Optional[str] = None,
        *,
        metadata: Optional[dict] = None
    ) -> str:
        """
        Creates a report

        Args:
            plots_analysis_id: id of the plots analysis
            report_name: name to give to the report
            plot_ids: list of the plot ids to select for the report
            report_type: type of report to generate, as per list_plots_analysis_report_types
            metadata:  set of key-value pairs which may be included in the report

        Deprecated arguments:
            plots_group_id: ignored, do not provide it

        Returns:
            str: the id of the new report
        """
        if plots_group_id is not None:
            warnings.warn("Passing plots_group_id is not needed anymore, remove it", DeprecationWarning)

        upload_id, upload_url = self._make_upload()
        resp = requests.put(upload_url, data=json.dumps({"plot_ids": plot_ids}))
        _check_resp_is_ok(resp, "Failure uploading plots file for analysis")
        data = {
            "name": report_name,
            "upload_id": upload_id,
            "report_type": report_type,
            "metadata": metadata if metadata is not None else {}
        }
        resp = self.sess.post(
            self._full_url(f"plots_analyses/{plots_analysis_id}/reports/"),
            json=data
        )
        _check_resp_is_ok(resp, "Failure starting analysis precheck")
        op_result = self._wait_until_operation_completes(resp.json())
        report_id = op_result["results"]["plots_analysis_report_id"]
        return report_id

    def get_plots_analysis(self, plots_analysis_id: str, plots_group_id: Optional[str] = None) -> Dict[str, Any]:
        """
        Get plots analysis information

        Args:
            plots_analysis_id: id of the plots analysis

        Deprecated arguments:
            plots_group_id: ignored, do not provide it

        Returns:
            dict: see https://app.picterra.ch/public/apidocs/plots_analysis/v1/#tag/analysis/operation/getAnalysis
        """
        if plots_group_id is not None:
            warnings.warn("Passing plots_group_id is not needed anymore, remove it", DeprecationWarning)

        resp = self.sess.get(self._full_url(f"plots_analyses/{plots_analysis_id}/"))
        _check_resp_is_ok(resp, "Failed to get plots analysis")
        return resp.json()

    def get_plots_analysis_report(self, plots_analysis_report_id: str, plots_group_id: Optional[str] = None, plots_analysis_id: Optional[str] = None)  -> Dict[str, Any]:
        """
        Get plots analysis report information

        Args:
            plots_analysis_report_id: id of the plots analysis report

        Deprecated arguments:
            plots_group_id: ignored, do not provide it
            plots_analysis_id: ignored, do not provide it

        Raises:
            APIError: There was an error while getting the plots analysis report information

        Returns:
            dict: see https://app.picterra.ch/public/apidocs/plots_analysis/v1/#tag/reports/operation/getReportForAnalysis
        """
        if plots_group_id is not None or plots_analysis_id is not None:
            warnings.warn("Passing plots_group_id/plots_analysis_id is not needed anymore, remove it", DeprecationWarning)

        resp = self.sess.get(self._full_url(f"plots_analysis_reports/{plots_analysis_report_id}/"))
        _check_resp_is_ok(resp, "Failed to get plots analysis report")
        return resp.json()



import datetime
import json
import os
import tempfile

import responses

from picterra import TracerClient
from picterra.base_client import multipolygon_to_polygon_feature_collection
from tests.utils import (
    OP_RESP,
    OPERATION_ID,
    _add_api_response,
    _client,
    add_mock_paginated_list_response,
    plots_analysis_api_url,
)


def make_geojson_polygon(base=1):
    return {"type": "Polygon", "coordinates": [[[0, 0], [base, 0], [base, base], [0, base], [0, 0]]]}


def make_geojson_multipolygon(npolygons=1):
    coords = []
    for i in range(npolygons):
        coords.append(make_geojson_polygon(i + 1)["coordinates"])
    return {"type": "MultiPolygon", "coordinates": coords}

def test_plots_analysis_platform_client_base_url(monkeypatch):
    """
    Sanity-check that the client defaults to the correct base url
    """
    monkeypatch.setenv("PICTERRA_API_KEY", "1234")
    client = TracerClient()
    assert client.base_url == "https://app.picterra.ch/public/api/plots_analysis/v1/"


@responses.activate
def test_create_plots_group(monkeypatch):
    _add_api_response(
        plots_analysis_api_url("upload/file/"),
        responses.POST,
        {
            "upload_id": "an-upload",
            "upload_url": "https://upload.example.com/",
        },
    )
    responses.put("https://upload.example.com/")
    _add_api_response(
        plots_analysis_api_url("plots_groups/"),
        responses.POST,
        OP_RESP,
        match=responses.matchers.json_params_matcher({
            "name": "name of my plot group",
            "methodology_id": "eudr-cocoa-id",
            "custom_columns_values": {"foo": "bar"}
        }),
    )
    _add_api_response(plots_analysis_api_url(f"operations/{OPERATION_ID}/"), responses.GET, {
        "status": "success",
        "results": {"plots_group_id": "a-plots-group"}
    })
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    with tempfile.NamedTemporaryFile() as tmp:
        _add_api_response(plots_analysis_api_url(
            "plots_groups/a-plots-group/upload/commit/"),
            responses.POST,
            OP_RESP,
            match=responses.matchers.json_params_matcher({
                "files": [
                    {
                        "upload_id": "an-upload",
                        "filename": os.path.basename(tmp.name),
                    }
                ],
                "overwrite": False,
            }),
        )
        with open(tmp.name, "w") as f:
            json.dump({"type": "FeatureCollection", "features": []}, f)
        assert client.create_plots_group(
            "name of my plot group",
            "eudr-cocoa-id",
            [tmp.name],
            {"foo": "bar"},
        ) == "a-plots-group"


@responses.activate
def test_update_plots_group_plots(monkeypatch):
    _add_api_response(
        plots_analysis_api_url("upload/file/"),
        responses.POST,
        {
            "upload_id": "an-upload",
            "upload_url": "https://upload.example.com/",
        },
    )
    responses.put("https://upload.example.com/")
    _add_api_response(plots_analysis_api_url(f"operations/{OPERATION_ID}/"), responses.GET, {
        "status": "success",
        "results": {
            "plots_group_id": "group-id",
        }
    })
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    with tempfile.NamedTemporaryFile() as tmp:
        _add_api_response(plots_analysis_api_url(
            "plots_groups/group-id/upload/commit/"),
            responses.POST,
            OP_RESP,
            match=responses.matchers.json_params_matcher({
                "files": [
                    {
                        "filename": os.path.basename(tmp.name),
                        "upload_id": "an-upload",
                    }
                ],
                "overwrite": False,
            }),
        )
        with open(tmp.name, "w") as f:
            json.dump({"type": "FeatureCollection", "features": []}, f)
        client.update_plots_group_plots("group-id", [tmp.name])


@responses.activate
def test_analyse_plots(monkeypatch):
    _add_api_response(
        plots_analysis_api_url("upload/file/"),
        responses.POST,
        {
            "upload_id": "an-upload-id",
            "upload_url": "https://upload.example.com/",
        },
    )
    responses.put("https://upload.example.com/", match=[responses.matchers.json_params_matcher({
        "plot_ids": ["uno", "dos"],
    })])
    _add_api_response(plots_analysis_api_url(
        "plots_groups/a-group-id/analysis/"),
        responses.POST,
        OP_RESP,
        match=responses.matchers.json_params_matcher({
            "analysis_name": "foobar",
            "upload_id": "an-upload-id",
            "date_from": "2023-01-01",
            "date_to": "2025-01-01",
        }),
    )
    _add_api_response(plots_analysis_api_url(f"operations/{OPERATION_ID}/"), responses.GET, {
        "status": "success",
        "results": {"analysis_id": "an-analysis-id"}
    })
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    with tempfile.NamedTemporaryFile() as tmp:
        with open(tmp.name, "w") as f:
            json.dump({"type": "FeatureCollection", "features": []}, f)
    assert client.analyze_plots(
        "a-group-id",
        "foobar",
        ["uno", "dos"],
        datetime.date.fromisoformat("2023-01-01"),
        datetime.date.fromisoformat("2025-01-01")
    ) == "an-analysis-id"


@responses.activate
def test_analyse_precheck(monkeypatch):
    _add_api_response(
        plots_analysis_api_url("upload/file/"),
        responses.POST,
        {
            "upload_id": "an-upload-id",
            "upload_url": "https://upload.example.com/",
        },
    )
    responses.put("https://upload.example.com/", match=[responses.matchers.json_params_matcher({
        "plot_ids": ["uno", "dos"],
    })])
    _add_api_response(plots_analysis_api_url(
        "plots_groups/a-group-id/analysis/precheck/"),
        responses.POST,
        OP_RESP,
        match=responses.matchers.json_params_matcher({
            "analysis_name": "foobar",
            "upload_id": "an-upload-id",
            "date_from": "2023-01-01",
            "date_to": "2025-01-01",
        }),
    )
    _add_api_response(plots_analysis_api_url(f"operations/{OPERATION_ID}/"), responses.GET, {
        "status": "success",
        "results": {"precheck_data_url": "https://precheck_data_url.example.com/"}
    })
    precheck = {
        "status": "failed",
        "errors": {"critical": [], "high": []},
        "critical_count": 1,
        "high_count": 1,
    }
    responses.get("https://precheck_data_url.example.com/", json=precheck)
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    with tempfile.NamedTemporaryFile() as tmp:
        with open(tmp.name, "w") as f:
            json.dump({"type": "FeatureCollection", "features": []}, f)
    assert client.analyze_plots_precheck(
        "a-group-id",
        "foobar",
        ["uno", "dos"],
        datetime.date.fromisoformat("2023-01-01"),
        datetime.date.fromisoformat("2025-01-01")
    ) == precheck


@responses.activate
def test_list_methodologies(monkeypatch):
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    url = plots_analysis_api_url("methodologies/")
    # Full list
    add_mock_paginated_list_response(url)
    methodologies = client.list_methodologies()
    assert len(methodologies) == 2
    assert methodologies[0]["name"] == "a_1"
    assert methodologies[1]["name"] == "a_2"
    # Search list
    add_mock_paginated_list_response(url, 2, "m_2", "spam")
    methodologies = client.list_methodologies(search="m_2", page_number=2)
    assert methodologies[0]["name"] == "spam_1"


@responses.activate
def test_list_plots_groups(monkeypatch):
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    url = plots_analysis_api_url("plots_groups/")
    # Full list
    add_mock_paginated_list_response(url)
    plots_groups = client.list_plots_groups()
    assert len(plots_groups) == 2
    assert plots_groups[0]["name"] == "a_1"
    assert plots_groups[1]["name"] == "a_2"
    # Search list
    add_mock_paginated_list_response(url, 2, "m_2", "spam")
    plots_groups = client.list_plots_groups(search="m_2", page_number=2)
    assert plots_groups[0]["name"] == "spam_1"


@responses.activate
def test_list_plots_analyses(monkeypatch):
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    url = plots_analysis_api_url("plots_groups/spam/analysis/")
    # Full list
    add_mock_paginated_list_response(url)
    plots_analyses = client.list_plots_analyses("spam")
    assert len(plots_analyses) == 2
    assert plots_analyses[0]["name"] == "a_1"
    assert plots_analyses[1]["name"] == "a_2"
    # Search list
    add_mock_paginated_list_response(url, 2, "m_2", "spam")
    plots_analyses = client.list_plots_analyses("spam", search="m_2", page_number=2)
    assert plots_analyses[0]["name"] == "spam_1"
    # Test the include_archived flag
    add_mock_paginated_list_response(url, include_archived=True, name_prefix="boo")
    plots_analyses = client.list_plots_analyses("spam", include_archived=True)
    assert len(plots_analyses) == 2
    assert plots_analyses[0]["name"] == "boo_1"


@responses.activate
def test_download_plots_group(monkeypatch):
    _add_api_response(plots_analysis_api_url(
        "plots_groups/a-group-id/export/"),
        responses.POST,
        OP_RESP,
        match=responses.matchers.json_params_matcher({"format": "geojson"}),
    )
    _add_api_response(plots_analysis_api_url(f"operations/{OPERATION_ID}/"), responses.GET, {
        "status": "success",
        "results": {"download_url": "https://a-group-id.example.com/geojson"}
    })
    polygons_fc = multipolygon_to_polygon_feature_collection(make_geojson_multipolygon())
    responses.add(
        responses.GET,
        "https://a-group-id.example.com/geojson",
        body=json.dumps(polygons_fc),
    )
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    with tempfile.NamedTemporaryFile() as f:
        client.download_plots_group_to_file("a-group-id", "geojson", f.name)
        assert json.load(f) == polygons_fc
    assert len(responses.calls) == 3


@responses.activate
def test_list_plots_analysis_reports(monkeypatch):
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    url = plots_analysis_api_url("plots_analyses/my-analysis-id/reports/")
    # Full list
    add_mock_paginated_list_response(url, num_results=3)
    reports = client.list_plots_analysis_reports("my-analysis-id")
    assert len(reports) == 3
    assert reports[0]["name"] == "a_1" and reports[-1]["name"] == "a_3"


@responses.activate
def test_list_plots_analysis_report_types(monkeypatch):
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    url = plots_analysis_api_url("plots_analyses/my-analysis-id/reports/types/")
    responses.get(
        url,
        json=[
            {"report_type": "type_1", "name": "a_1"},
            {"report_type": "type_2", "name": "a_2"},
            {"report_type": "type_1", "name": "a_3"},
            {"report_type": "type_3", "name": "a_4"},
        ],
    )
    reports = client.list_plots_analysis_report_types("my-analysis-id")
    assert len(reports) == 4
    assert reports[0]["report_type"] == "type_1" and reports[-1]["name"] == "a_4"


@responses.activate
def test_create_plots_analysis_report_precheck(monkeypatch):
    _add_api_response(
        plots_analysis_api_url("upload/file/"),
        responses.POST,
        {
            "upload_id": "an-upload-id",
            "upload_url": "https://upload.example.com/",
        },
    )
    responses.put("https://upload.example.com/", match=[responses.matchers.json_params_matcher({
        "plot_ids": ["uno", "dos"],
    })])
    _add_api_response(plots_analysis_api_url(
        "plots_analyses/an-analysis-id/reports/precheck/"),
        responses.POST,
        OP_RESP,
        match=responses.matchers.json_params_matcher({
            "name": "foobar",
            "upload_id": "an-upload-id",
            "report_type": "a-report-type",
            "metadata": {"foo": "bar"},
        }),
    )
    _add_api_response(plots_analysis_api_url(f"operations/{OPERATION_ID}/"), responses.GET, {
        "status": "success",
        "results": {"precheck_data_url": "https://precheck_data_url.example.com/"}
    })
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    with tempfile.NamedTemporaryFile() as tmp:
        with open(tmp.name, "w") as f:
            json.dump({"type": "FeatureCollection", "features": []}, f)
    assert client.create_plots_analysis_report_precheck(
        "an-analysis-id",
        "foobar",
        ["uno", "dos"],
        "a-report-type",
        metadata={"foo": "bar"}
    ) == {"status": "passed"}


@responses.activate
def test_create_plots_analysis_report(monkeypatch):
    _add_api_response(
        plots_analysis_api_url("upload/file/"),
        responses.POST,
        {
            "upload_id": "an-upload-id",
            "upload_url": "https://upload.example.com/",
        },
    )
    responses.put("https://upload.example.com/", match=[responses.matchers.json_params_matcher({
        "plot_ids": ["uno", "dos"],
    })])
    _add_api_response(plots_analysis_api_url(
        "plots_analyses/an-analysis-id/reports/"),
        responses.POST,
        OP_RESP,
        match=responses.matchers.json_params_matcher({
            "name": "foobar",
            "upload_id": "an-upload-id",
            "report_type": "a-report-type",
            "metadata": {"foo": "bar"},
        }),
    )
    _add_api_response(plots_analysis_api_url(f"operations/{OPERATION_ID}/"), responses.GET, {
        "status": "success",
        "results": {"plots_analysis_report_id": "a-report-id"}
    })
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    with tempfile.NamedTemporaryFile() as tmp:
        with open(tmp.name, "w") as f:
            json.dump({"type": "FeatureCollection", "features": []}, f)
    assert client.create_plots_analysis_report(
        "an-analysis-id",
        "foobar",
        ["uno", "dos"],
        "a-report-type",
        metadata={"foo": "bar"}
    ) == "a-report-id"


@responses.activate
def test_get_plots_analysis(monkeypatch):
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    _add_api_response(
        plots_analysis_api_url("plots_analyses/an-analysis-id/"),
        responses.GET,
        {
            "id": "an-analysis-id",
            "name": "My Analysis",
            "date_from": "2023-06-06",
            "date_to": "2025-02-08",
            "url": "https://app.picterra.ch/plots_analysis/plots_groups/136b812e-8d9c-418f-b317-8be5c7c6281d/analysis/cda443d7-5baf-483d-bb5e-fa1190180b0d/"  # noqa[E501]
        },
    )
    plots_analysis = client.get_plots_analysis("an-analysis-id")
    assert plots_analysis["id"] == "an-analysis-id"
    assert plots_analysis["name"] == "My Analysis"


@responses.activate
def test_get_plots_analysis_report(monkeypatch):
    _add_api_response(
        plots_analysis_api_url(
            "plots_analysis_reports/a-report-id/"
        ),
        responses.GET,
        {
            "id": "a-report-id",
            "name": "my report",
            "created_at": "2025-09-29T10:04:08.143098Z",
            "report_type": "eudr_export",
            "artifacts": [
                {
                    "name": "EUDR Report",
                    "filename": "2025-09-29-nightly-eudr-export.pdf",
                    "size_bytes": 71802,
                    "description": "A PDF report to be used for EUDR",
                    "content_type": "application/pdf",
                    "download_url": "http://example.com/report.pdf",
                },
                {
                    "name": "EUDR Traces NT",
                    "filename": "2025-09-29-nightly-eudr-export.geojson",
                    "size_bytes": 877,
                    "description": "A GeoJSON file that can be submitted to the EU Deforestation Due Diligence Registry",
                    "content_type": "application/geo+json",
                    "download_url": "http://example.com/traces.geojson",
                },
            ],
        },
    )
    client: TracerClient = _client(monkeypatch, platform="plots_analysis")
    report = client.get_plots_analysis_report("a-report-id")
    assert report["id"] == "a-report-id"
    assert report["artifacts"][0]["name"] == "EUDR Report"
